///////////////////////////////////////////////////////////////////////////////////
//
// Functions mainly used for development and debugging
//
///////////////////////////////////////////////////////////////////////////////////
// In the final version most of this code will be commented out and not compiled

void DebugReport(int zRPnn ) {
  ///////////////////////////////////////////////////////////////////////
  // a choice of print statements to inspect variables dynamically
  ///////////////////////////////////////////////////////////////////////
  // note at 4ms cycle and 115200 baud string length is limited to 46 chars
  // max in a 4ms period, otherwise blocking will occur. If used of WiFi then
  // link corruption may occur; so to prevent rogue values we send data twice and
  // checked on receive as a valid pair of duplicate values, otherwise disgard it.
  switch (zRPnn) {
    case 0:
      if (RP_Title) {PrintTx += "#AccRawY,AccRawZ,0\n"; RP_Title = false;}
      PrintTx += String(AccRawY) + ",";
      PrintTx += String(AccRawZ) + ",";
      PrintTx += "0\n";
      break;
      
    case 1:
      if (RP_Title) {PrintTx += "#angle_acc,angle_gyro,angle_diff,0\n"; RP_Title = false;}
      PrintTx += String(angle_acc) + ",";
      PrintTx += String(angle_gyro) + ",";
      PrintTx += String(angle_acc - angle_gyro) + ",";
      PrintTx += "0\n";
      break;
      
    case 2:
      if (RP_Title) {PrintTx += "#GyrRawX,GyrRawY,GyrRawZ,0\n"; RP_Title = false;}
      PrintTx += String(gyro_pitch_data_raw) + ",";
      PrintTx += String(gyro_turn_data_raw) + ",";
      PrintTx += String(gyro_yaw_data_raw) + ",";
      PrintTx += "0\n";
      break;
      
    case 3:
      if (RP_Title) {PrintTx += "#PID_p,PID_i,PID_d,PID o/p\n"; RP_Title = false;}
      PrintTx += String(PID_p) + ",";
      PrintTx += String(pid_i_mem) + ",";
      PrintTx += String(PID_d) + ",";
      PrintTx += String(pid_output) + "\n";
      break;

    case 4:
      if (RP_Title) {PrintTx += "#RangeRaw,Range,0,500\n"; RP_Title = false;}
      PrintTx += String(RangeRaw) + ",";
      PrintTx += String(Range) + ",";
      PrintTx += "0,500\n";
      break;

    case 5:
      if (RP_Title) {PrintTx += "#RpmLft,RpmRht,0\n"; RP_Title = false;}
      PrintTx += String(Int34RPM) + ",";
      PrintTx += String(Int35RPM) + ",";
      PrintTx += "0\n";
      break;

    case 6:
      if (RP_Title) {PrintTx += "#DIR_LTnd,DIR_RTnd,0\n"; RP_Title = false;}
      PrintTx += String(DIR_LTnd) + ",";
      PrintTx += String(DIR_RTnd) + ",";
      PrintTx += "0\n";
      break;

    case 7:
      if (RP_Title) {PrintTx += "#LftPeriod,RhtPeriod,0\n"; RP_Title = false;}
      PrintTx += String(IntPeriods[0]) + ",";
      PrintTx += String(IntPeriods[1]) + ",";
      PrintTx += "0\n";
      break;


    default:
      // zRPnn was out of range so reset the reporting pointer and send commands
      // to set RP = 0 and RMax to the highest number in your case list. ie. 7
      RPnn = 0; PrintTx += "RP0\n";  PrintTx += "RM7\n"; break;
  }
}

// --------------------------------------------------------------------------------

void cmdReset() {
  // turn-OFF drive motors in case they were set in motion
//  driveMotors(0.0, 0.0, 0.0, 0.0);

  // recalibrate motor PWM mapping  
//  MotorCal = true;    // controls motor PWM calibration, default = true
//  motorCal();         // invoke PWM remapping
//
//  // resets DEBUG variables to the default condition  
//  cmdMode = ' ';      // command mode
//  cmdRec;             // > 0 if a '~' has been received
//  cmdSgn = 1;         // value polarity multiplier
//  cmdSteer = 0;       // values used in motor steering
//  cmdType = ' ';      // command type
//  cmdVal = 0;         // value associated with a cmdType
//  cmdXDrive = 0;      // values used in motor testing
//  cmdYDrive = 0;      // values used in motor testing
//  keyChar;            // any keyboard character
//  keyVal;             // any keyboard value
//  PID_SPJX = false;   // PID setpoint X jolt flag
//  PID_SPJY = false;   // PID setpoint Y jolt flag
//  RP_ON = false;      // if true then report values in DEBUG mode
//  Steer_En = false;   // steering enabled flag
//  X_DriveEn = false;  // motor X-drive enabled flag
//  Y_DriveEn = false;  // motor Y-drive enabled flag
//
//  Serial.println("!"); // '!' echo
}

// --------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received so execute command if valid
  // supported commands are:
  //
  //    D0.   - disable PID d-gain
  //    D1.   - enable PID d-gain
  //    DD.   - decrease the d-gain by 0.1
  //    DU.   - increase the d-gain by 0.1
  //    DVnn. - set 'D' gain value to nn/10
  //    I0.   - disable PID i-gain
  //    I1.   - enable PID i-gain
  //    ID.   - decrease the i-gain by 0.1
  //    IU.   - increase the i-gain by 0.1
  //    IVnn. - set 'I' gain value to nn/100
  //    JXnn. - set global variable JX to nn, used for debug purposes
  //    JYnn. - set global variable JY to nn, used for debug purposes
  //    PD.   - decrease the P-gain by 1.0
  //    PU.   - increase the P-gain by 1.0
  //    PVnn. - set 'P' gain value to nn/10
  //    R.    - report over the serial link
  //    RPnn. - report nn data over the serial link
  //    SGnn. - set 'S' gain value to nn
  //    STn.   - set servo target to 'n'
  //    SVnn.  - set value of servoPnt device to nn
  //  .     - report the P,I,D gains and much more!
  
  bool zPLF = true;   // if true at the end a line feed will be printed
  bool zErr = false;  // if a command is not recognised it simply echoes it
  switch (cmdMode) {
    case ' ':
      // recevied a '.' only so report PID gains
      // block this if in graphing mode as it could be data corruption
      if (!RP_ON) {PrintTx += "\n"; printGains(); zPLF = false;}
      break;
    case 'D':
      // PID controller D commands
      switch (cmdType) {
        case ' ': 
          if (cmdVal > 0) {PID_d_En = true;}
          else {PID_d_En = false;} break;
        case 'D': pid_d_gain -= 0.1;  break;
        case 'U': pid_d_gain += 0.1; break;
        case 'V': pid_d_gain = (float)cmdVal/10.0; break;
        default: zErr = true; break;
      } 
      PrintTx += "D" + String((int)(pid_d_gain * 10.0)); break;
    case 'I':
      // PID controller I commands
      switch (cmdType) {
        case ' ': 
          if (cmdVal > 0) {PID_i_En = true;}
          else {PID_i_En = false;} break;
        case 'D': pid_i_gain -= 0.1;  break;
        case 'U': pid_i_gain += 0.1; break;
        case 'V': pid_i_gain = (float)cmdVal/100.0; break;
        default: zErr = true; break;
      }
      if (pid_i_gain < 0.01) {pid_i_mem = 0.0;}
      PrintTx +="I" + String((int)(pid_i_gain * 100.0)); break;
    case 'J':
      // commands used to set global variables in association with Joystick app
      switch (cmdType) {
        case 'X': // set global variable JX to cmdVal
          JX = cmdVal; //Serial.println("JX=" + String(JX));
          break;
        case 'Y': // set global variable JY to cmdVal
          JY = cmdVal; //Serial.println("JY=" + String(JY));
          break;
        default: zErr = true; break;
      } break;
    case 'P':
      // PID controller P commands
      switch (cmdType) {
        case 'D': pid_p_gain -= 1.0;  break;
        case 'U': pid_p_gain += 1.0; break;
        case 'V': pid_p_gain = (float)cmdVal/10.0; break;
        default: zErr = true; break;
      }
      PrintTx += "P" + String((int)(pid_p_gain * 10.0)); break;
    case 'R':
      // PID controller report commands
      switch (cmdType) {
        case ' ': 
          // report the same set of values twice as WiFi data can easily be
          // corrupted. Therefore two matching consequitives messages must be
          // received for the PID Controller graphing function to work
          RP_ON = true; // set flag to indicate that we are reporting data
          RPms = millis() - RPt0; RPt0 = millis();
          RPP = 0; DebugReport(RPnn);
          RPP = 1; DebugReport(RPnn);
          zPLF = false; break;
        case 'E': // RE. exits reporting mode when dropping out of graphing mode
          RP_ON = false;
          break;
        case 'P': // RPnn. sets the reporting mode and triggers title reporting
          RPnn = cmdVal; RP_Title = true;
          PrintTx += "RP" + String(RPnn) + "\n"; break;
        default: zErr = true; break;
      } break;
    case 'S':
      // Servo & Motor Steer commands & PIS setpoint jolt
      switch (cmdType) {
        case 'G':
          PWM_Start = (float)cmdVal;
          PrintTx += "SG"+String(cmdVal); break;
        case 'T':
          ServoPnt = cmdVal;
          PrintTx += "ST"+String(cmdVal);
          break;
        case 'V':
          if (ServoPnt == 0) {
            // this is the head servo
            cmdVal = constrain(cmdVal,500,2400); HeadVal = cmdVal;
            if (!HeadAtt) {attachHeadServo();}
            else {servoHead.writeMicroseconds(HeadVal); HeadTgt = HeadVal;}
          } else {
            // this is the left motor
            cmdVal = cmdVal * cmdSgn;
            cmdVal = constrain(cmdVal,-255,255);
            if (ServoPnt == 1) {PWM_Lft = cmdVal;}
            if (ServoPnt == 2) {PWM_Rht = cmdVal;}
          }
          PrintTx += "SV"+String(cmdVal); break;
        default: zErr = true; break;
      } break;
    default: zErr = true; break;
  }
  // echo unrecognised commands
  if (zErr) {PrintTx += String(cmdMode) + String(cmdType) + String(cmdVal);}
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
  if (zPLF) {PrintTx += "\n";}
}

// --------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = (cmdVal * 10) + zVal;  
}

// --------------------------------------------------------------------------------

void FpSerialPrint(float zP) {
  // print floating point number limited to 4 significant digits, without CR
  if (abs(zP) < 100.0) {Serial.print(zP); return;}
  if (abs(zP) < 1000.0) {Serial.print(zP,1); return;}
  Serial.print((int)zP);
}

// --------------------------------------------------------------------------------

void FpSerialPrintln(float zP) {
  // print floating point number limited to 4 significant digits, with CR
  if (abs(zP) < 100.0) {Serial.println(zP); return;}
  if (abs(zP) < 1000.0) {Serial.println(zP,1); return;}
  Serial.println((int)zP);
}

// --------------------------------------------------------------------------------

void printGains() {
  // print the current gain factors
  // this is used by the Windows PID controller app to load the current gain settings
  PrintTx += "PID=" + String(pid_p_gain) + ",\t";
  if (PID_i_En) {PrintTx += String(pid_i_gain);}
  else {PrintTx += "--.--";}
  PrintTx += ",\t";
  if (PID_d_En) {PrintTx += String(pid_d_gain);}
  else {PrintTx += "--.--";}
  PrintTx += ",\t";
  PrintTx += String(PWM_Start) + "\n";
}

// --------------------------------------------------------------------------------
